# coding=utf-8
#
# Copyright © 2013, 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""Generate interpolation-qualifier tests."""

from __future__ import print_function, division, absolute_import
import os
import itertools

from templates import template_dir
from modules import utils

TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))

INTERPOLATION_MODES = [
    'flat',
    'noperspective',
    'smooth',
    'default'
]

VS_VARIABLES = [
    'gl_FrontColor',
    'gl_BackColor',
    'gl_FrontSecondaryColor',
    'gl_BackSecondaryColor'
]

VS_VARIABLES_FRONT_ONLY = [
    'gl_FrontColor',
    'gl_FrontSecondaryColor',
]

OTHER_SIDE_MAP = {
    'gl_FrontColor': 'gl_BackColor',
    'gl_BackColor': 'gl_FrontColor',
    'gl_FrontSecondaryColor': 'gl_BackSecondaryColor',
    'gl_BackSecondaryColor': 'gl_FrontSecondaryColor'
}

VS_TO_FS_VARIABLE_MAP = {
    'gl_FrontColor': 'gl_Color',
    'gl_BackColor': 'gl_Color',
    'gl_FrontSecondaryColor': 'gl_SecondaryColor',
    'gl_BackSecondaryColor': 'gl_SecondaryColor'
}


def make_fs_vs_tests(fs_mode, vs_mode, dirname):
    if vs_mode == fs_mode:
        return

    for var in VS_VARIABLES:
        filename = os.path.join(
            dirname,
            '{0}-{1}-{2}-{3}.shader_test'.format(
                vs_mode, var, fs_mode,
                VS_TO_FS_VARIABLE_MAP[var]))
        print(filename)

        with open(filename, 'w') as f:
            f.write(TEMPLATES.get_template('vs-fs.shader_test.mako').render_unicode(
                vs_mode=vs_mode,
                vs_variable=var,
                fs_mode=fs_mode,
                fs_variable=VS_TO_FS_VARIABLE_MAP[var]))


def make_vs_unused_tests(vs_mode, dirname):
    if vs_mode == 'default':
        return

    for var in VS_VARIABLES:
        filename = os.path.join(
            dirname,
            '{0}-{1}-unused-{2}.shader_test'.format(
                vs_mode, var,
                VS_TO_FS_VARIABLE_MAP[var]))
        print(filename)

        with open(filename, 'w') as f:
            f.write(
                TEMPLATES.get_template('vs-unused.shader_test.mako').render_unicode(
                    vs_mode=vs_mode,
                    vs_variable=var))


def make_fs_unused_tests(fs_mode, vs_mode, dirname):
    if fs_mode == 'default':
        return

    for var in VS_VARIABLES_FRONT_ONLY:
        filename = os.path.join(
            dirname,
            'unused-{0}-{1}-{2}.shader_test'.format(
                var, fs_mode,
                VS_TO_FS_VARIABLE_MAP[var]))
        print(filename)

        with open(filename, 'w') as f:
            f.write(TEMPLATES.get_template('fs-unused.shader_test.mako').render_unicode(
                vs_mode=vs_mode,
                vs_variable=var,
                fs_mode=fs_mode,
                fs_variable=VS_TO_FS_VARIABLE_MAP[var]))


def make_vs_fs_unused_tests(fs_mode, vs_mode, dirname):
    if vs_mode == fs_mode:
        return

    for var in VS_VARIABLES:
        filename = os.path.join(
            dirname,
            'unused-{0}-{1}-unused-{2}-{3}.shader_test'.format(
                vs_mode, var, fs_mode,
                VS_TO_FS_VARIABLE_MAP[var]))
        print(filename)

        with open(filename, 'w') as f:
            f.write(TEMPLATES.get_template(
                'fs-vs-unused.shader_test.mako').render_unicode(
                    vs_mode=vs_mode,
                    vs_variable=var,
                    fs_mode=fs_mode,
                    fs_variable=VS_TO_FS_VARIABLE_MAP[var]))


def make_vs_fs_flip_tests(fs_mode, vs_mode, dirname):
    if vs_mode == fs_mode:
        return

    for this_side in VS_VARIABLES:
        other_side = OTHER_SIDE_MAP[this_side]
        filename = os.path.join(
            dirname,
            '{0}-{1}-{2}-{3}.shader_test'.format(
                vs_mode, this_side, fs_mode, other_side))
        print(filename)

        with open(filename, 'w') as f:
            f.write(TEMPLATES.get_template(
                'vs-fs-flip.shader_test.mako').render_unicode(
                    vs_mode=vs_mode,
                    this_side_variable=this_side,
                    other_side_variable=other_side,
                    fs_mode=fs_mode,
                    fs_variable=VS_TO_FS_VARIABLE_MAP[this_side]))


def main():
    """main function."""
    dirname = os.path.join('spec', 'glsl-1.30', 'linker',
                           'interpolation-qualifiers')
    utils.safe_makedirs(dirname)

    for fs_mode, vs_mode in itertools.product(INTERPOLATION_MODES, repeat=2):
        make_fs_vs_tests(fs_mode, vs_mode, dirname)
        make_vs_unused_tests(vs_mode, dirname)
        make_fs_unused_tests(fs_mode, vs_mode, dirname)
        make_vs_fs_unused_tests(fs_mode, vs_mode, dirname)
        make_vs_fs_flip_tests(fs_mode, vs_mode, dirname)


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright 2013 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Authors: Tom Stellard <thomas.stellard@amd.com>
#
#

from __future__ import print_function, division, absolute_import
import os
import textwrap

from six.moves import range

from modules import utils

TYPES = ['char', 'uchar', 'short', 'ushort', 'int', 'uint', 'long', 'ulong', 'float', 'double']
VEC_SIZES = ['', '2', '4', '8', '16']

dirName = os.path.join("cl", "store")


def gen_array(size):
    return ' '.join([str(i) for i in range(size * 8)])


def ext_req(type_name):
    if type_name[:6] == "double":
        return "require_device_extensions: cl_khr_fp64"
    return ""


def print_config(f, type_name, addr_space):
    f.write(textwrap.dedent(("""
    [config]
    name: Store {type_name}
    program_source_file: store-kernels-{addr_space}.inc
    build_options: -D TYPE={type_name}
    dimensions: 1
    """ + ext_req(type_name))
    .format(type_name=type_name, addr_space=addr_space)))


def begin_test(type_name, addr_space):
    fileName = os.path.join(dirName, 'store-' + type_name + '-' + addr_space + '.program_test')
    print(fileName)
    f = open(fileName, 'w')
    print_config(f, type_name, addr_space)
    return f


def main():
    utils.safe_makedirs(dirName)

    for t in TYPES:
        for s in VEC_SIZES:
            if s == '':
                size = 1
            else:
                size = int(s)
            type_name = t + s
            f = begin_test(type_name, 'global')
            f.write(textwrap.dedent("""
            [test]
            name: global address space
            global_size: 1 0 0
            kernel_name: store_global
            arg_out: 0 buffer {type_name}[8] {gen_array}
            arg_in:  1 buffer {type_name}[8] {gen_array}
            [test]
            name: global address space work items
            global_size: 8 0 0
            kernel_name: store_global_wi
            arg_out: 0 buffer {type_name}[8] {gen_array}
            arg_in:  1 buffer {type_name}[8] {gen_array}
            """.format(type_name=type_name, gen_array=gen_array(size))))

            f.close()

            f = begin_test(type_name, 'local')
            f.write(textwrap.dedent("""
            [test]
            name: local address space
            global_size: 8 0 0
            local_size:  8 0 0
            kernel_name: store_local
            arg_out: 0 buffer {type_name}[8] {gen_array}
            arg_in:  1 buffer {type_name}[8] {gen_array}
            """.format(type_name=type_name, gen_array=gen_array(size))))

            f.close()


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright (c) 2014, 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Provides test for the framework.profile modules """

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import pytest
import six

from framework import exceptions
from framework import grouptools
from framework import profile
from . import utils

# pylint: disable=invalid-name,no-self-use,protected-access


class TestLoadTestProfile(object):
    """Tests for profile.load_test_profile."""

    def test_no_profile_attribute(self, tmpdir):
        """profile.load_test_profile: Loading a module with no profile name
        exits.

        Because load_test_profile uses test.{} to load a module we need a
        module in tests that doesn't have a profile attribute. The only module
        that currently meets that requirement is __init__.py
        """
        p = tmpdir.join('foo.profile')

        with pytest.raises(exceptions.PiglitFatalError):
            profile.load_test_profile(six.text_type(p))

    def test_no_such_module(self, tmpdir):
        """profile.load_test_profile: Trying to load a non-existent module
        exits.
        """
        # Ensure that the module will not exist by moving to an empty temporary
        # directory
        tmpdir.chdir()
        with pytest.raises(exceptions.PiglitFatalError):
            profile.load_test_profile('this_module_will_never_ever_exist')


class TestTestProfile(object):
    """Tests for profile.TestProfile."""

    class TestCopy(object):
        """Tests for the copy method."""

        @pytest.fixture
        def fixture(self):
            orig = profile.TestProfile()
            orig.test_list['foo'] = utils.Test(['foo'])
            orig.test_list['bar'] = utils.Test(['bar'])
            orig.filters = [lambda name, _: name != 'bar']
            orig.forced_test_list = ['foo']
            return orig

        def test_filters(self, fixture):
            """The filters attribute is copied correctly."""
            new = fixture.copy()

            # Assert that the fixtures are equivalent, but not the same
            assert fixture.filters == new.filters
            assert fixture.filters is not new.filters

            # And double check by modifying one of them and asserting that the
            # other has not changed.
            new.filters.append(lambda name, _: name != 'oink')
            assert len(fixture.filters) == 1

        def test_forced_test_list(self, fixture):
            """The forced_test_list attribute is copied correctly."""
            new = fixture.copy()

            # Assert that the fixtures are equivalent, but not the same
            assert fixture.forced_test_list == new.forced_test_list
            assert fixture.forced_test_list is not new.forced_test_list

            # And double check by modifying one of them and asserting that the
            # other has not changed.
            del new.forced_test_list[0]
            assert fixture.forced_test_list[0] == 'foo'

        def test_test_list(self, fixture):
            """The test_list attribute is copied correctly."""
            new = fixture.copy()

            # Assert that the fixtures are equivalent, but not the same
            assert fixture.test_list == new.test_list
            assert fixture.test_list is not new.test_list


class TestTestDict(object):
    """Tests for the TestDict object."""

    test = None

    def setup(self):
        self.test = profile.TestDict()

    @pytest.mark.parametrize(
        "arg",
        [None, utils.Test(['foo']), ['a'], {'a': 1}],
        ids=six.text_type)
    def test_key_not_string(self, arg):
        """profile.TestDict: If key value isn't a string an exception is raised.

        This throws a few different things at the key value and expects an
        error to be raised. It isn't a perfect test, but it was the best I
        could come up with.
        """
        with pytest.raises(exceptions.PiglitFatalError):
            self.test[arg] = utils.Test(['foo'])

    def test_reassignment(self):
        """reassigning a key raises an exception."""
        self.test['foo'] = utils.Test(['foo'])
        with pytest.raises(exceptions.PiglitFatalError):
            self.test['foo'] = utils.Test(['foo', 'bar'])

    class TestAllowReassignment(object):
        """Tests for TestDict.allow_reassignment."""

        @pytest.fixture
        def test(self):
            return profile.TestDict()

        def test_case_insensitve(self, test):
            """reassigning a key raises an exception (capitalization is
            ignored)."""
            test['foo'] = utils.Test(['foo'])
            with pytest.raises(exceptions.PiglitFatalError):
                test['Foo'] = utils.Test(['foo', 'bar'])

        def test_simple(self, test):
            """profile.TestDict.allow_reassignment works."""
            test['a'] = utils.Test(['foo'])
            with test.allow_reassignment:
                test['a'] = utils.Test(['bar'])

            assert test['a'].command == ['bar']

        def test_with_groupmanager(self, test):
            """profile.TestProfile: allow_reassignment wrapper works with
            groupmanager.
            """
            testname = grouptools.join('a', 'b')
            test[testname] = utils.Test(['foo'])
            with test.allow_reassignment:
                with test.group_manager(utils.Test, 'a') as g:
                    g(['bar'], 'b')

            assert test[testname].command == ['bar']

        def test_stacked(self, test):
            """profile.profile.TestDict.allow_reassignment: check stacking
            cornercase.

            There is an odd corner case in the original (obvious)
            implementation of this function, If one opens two context managers
            and then returns from the inner one assignment will not be allowed,
            even though one is still inside the first context manager.
            """
            test['a'] = utils.Test(['foo'])
            with test.allow_reassignment:
                with test.allow_reassignment:
                    pass
                test['a'] = utils.Test(['bar'])

            assert test['a'].command == ['bar']

    class TestGroupManager(object):
        """Test the group_manager method."""

        @pytest.fixture
        def inst(self):
            return profile.TestDict()

        def test_no_name_args_eq_one(self, inst):
            """no name and len(args) == 1 is valid."""
            with inst.group_manager(utils.Test, 'foo') as g:
                g(['a'])

            assert grouptools.join('foo', 'a') in inst

        def test_no_name_args_gt_one(self, inst):
            """no name and len(args) > 1 is valid."""
            with inst.group_manager(utils.Test, 'foo') as g:
                g(['a', 'b'])

            assert grouptools.join('foo', 'a b') in inst

        def test_name(self, inst):
            """name plus len(args) > 1 is valid."""
            with inst.group_manager(utils.Test, 'foo') as g:
                g(['a', 'b'], 'a')

            assert grouptools.join('foo', 'a') in inst

        def test_adder_kwargs_passed(self, inst):
            """Extra kwargs passed to the adder function are passed to the
            Test.
            """
            with inst.group_manager(utils.Test, 'foo') as g:
                g(['a'], run_concurrent=True)
            test = inst[grouptools.join('foo', 'a')]

            assert test.run_concurrent is True

        def test_context_manager_keyword_args_passed(self, inst):
            """kwargs passed to the context_manager are passed to the Test."""
            with inst.group_manager(
                    utils.Test, 'foo', run_concurrent=True) as g:  # pylint: disable=bad-continuation
                    # This is a pylint bug, there's an upstream report
                g(['a'])
            test = inst[grouptools.join('foo', 'a')]

            assert test.run_concurrent is True

        def test_adder_kwargs_overwrite_context_manager_kwargs(self, inst):
            """default_args are overwritten by kwargs."""
            with inst.group_manager(
                    utils.Test, 'foo', run_concurrent=True) as g:  # pylint: disable=bad-continuation
                    # This is a pylint bug, there's an upstream report
                g(['a'], run_concurrent=False)

            test = inst[grouptools.join('foo', 'a')]
            assert test.run_concurrent is False


class TestRegexFilter(object):
    """Tests for the RegexFilter class."""

    class TestNormal(object):
        """Tests for inverse set to False (default)."""

        def test_empty(self):
            """Returns True when no filters are provided."""
            test = profile.RegexFilter([])
            assert test('foobob', None)

        def test_matches(self):
            """Returns True when the test matches any regex."""
            test = profile.RegexFilter([r'foo', r'bar'])
            assert test('foobob', None)

        def test_not_matches(self):
            """Returns True when the test matches any regex."""
            test = profile.RegexFilter([r'fob', r'bar'])
            assert not test('foobob', None)

    class TestInverse(object):
        """Tests for inverse set to True."""

        def test_empty(self):
            """Returns True when no filters are provided."""
            test = profile.RegexFilter([], inverse=True)
            assert test('foobob', None)

        def test_matches(self):
            """Returns False when the test matches any regex."""
            test = profile.RegexFilter([r'foo', r'bar'], inverse=True)
            assert not test('foobob', None)

        def test_not_matches(self):
            """Returns False when the test matches any regex."""
            test = profile.RegexFilter([r'fob', r'bar'], inverse=True)
            assert test('foobob', None)



# coding=utf-8
# Copyright (c) 2014 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Generate spec/ARB_shader_texture_lod tests """

from __future__ import print_function, division, absolute_import
import os
import collections

from templates import template_dir
from modules import utils

TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))

Parameters = collections.namedtuple(
    'Parameters', ['coord', 'grad', 'dimensions', 'mode'])


LOD_TESTS = [
    Parameters('float', 'float', '1D', 'texture1D'),
    Parameters('vec2', 'float', '1D', 'texture1DProj'),
    Parameters('vec4', 'float', '1D', 'texture1DProj'),
    Parameters('vec3', 'float', '1DShadow', 'shadow1D'),
    Parameters('vec4', 'float', '1DShadow', 'shadow1DProj'),
    Parameters('vec2', 'vec2', '2D', 'texture2D'),
    Parameters('vec3', 'vec2', '2D', 'texture2DProj'),
    Parameters('vec4', 'vec2', '2D', 'texture2DProj'),
    Parameters('vec3', 'vec2', '2DShadow', 'shadow2D'),
    Parameters('vec4', 'vec2', '2DShadow', 'shadow2DProj'),
    Parameters('vec3', 'vec3', '3D', 'texture3D'),
    Parameters('vec4', 'vec3', '3D', 'texture3DProj'),
    Parameters('vec3', 'vec3', 'Cube', 'textureCube')
]

GRAD_TESTS = LOD_TESTS + [
    Parameters('vec2', 'vec2', '2DRect', 'texture2DRect'),
    Parameters('vec3', 'vec2', '2DRect', 'texture2DRectProj'),
    Parameters('vec3', 'vec2', '2DRectShadow', 'shadow2DRect'),
    Parameters('vec4', 'vec2', '2DRectShadow', 'shadow2DRectProj')
]


def get_extensions(mode):
    """ If this test uses GL_ARB_texture_rectangle add it

    GL_ARB_texture_rectangle is an odd extension, it is on by default, so don't
    generate a #extension in the shader, just in the config block.

    """
    if 'Rect' in mode:
        return 'GL_ARB_shader_texture_lod GL_ARB_texture_rectangle'
    return 'GL_ARB_shader_texture_lod'


def main():
    """ Main function

    Writes tests to generated_tests/spec/arb_shader_texture_lod/ directory

    """
    dirname = 'spec/arb_shader_texture_lod/compiler'
    utils.safe_makedirs(dirname)

    for params in LOD_TESTS:
        name = os.path.join(
            dirname,
            "tex_lod-{mode}-{dimensions}-{coord}.frag".format(
                mode=params.mode,
                dimensions=params.dimensions,
                coord=params.coord))
        print(name)
        with open(name, 'w+') as f:
            f.write(TEMPLATES.get_template(
                'frag_lod.glsl_parser_test.mako').render_unicode(param=params))

    for params in GRAD_TESTS:
        # Generate fragment shader test
        name = os.path.join(
            dirname,
            "tex_grad-{mode}-{dimensions}-{coord}".format(
                mode=params.mode,
                dimensions=params.dimensions,
                coord=params.coord))

        for stage in ['frag', 'vert']:
            print('{0}.{1}'.format(name, stage))
            with open('{0}.{1}'.format(name, stage), 'w+') as f:
                f.write(TEMPLATES.get_template(
                    'tex_grad.{0}.mako'.format(stage)).render_unicode(
                        param=params,
                        extensions=get_extensions(params.mode)))


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Copyright © 2018 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import print_function, division, absolute_import
import struct
import os
import numpy as np
import random

import six

from templates import template_file
from modules import utils

def generate_results_commutative(srcs, operator):
    """Generate results for an operator that is commutative.

    Commutative operators will only generate an upper-right triangular
    matrix of results, and the diagonal will be missing.
    """
    results = []
    for i in range(len(srcs) - 1):
        for j in range(i + 1, len(srcs)):
            results.append(operator(srcs[i], srcs[j]))

    return results


def generate_results_commutative_with_diagonal(srcs, operator):
    """Generate results for an operator that is commutative.

    Commutative operators will only generate an upper-right triangular
    matrix of results, but the diagonal must also be explicitly stored.
    """
    results = []
    for i in range(len(srcs)):
        for j in range(i, len(srcs)):
            results.append(operator(srcs[i], srcs[j]))

    return results


def generate_results_without_diagonal(srcs, operator):
    """Generate full matrix of results without the diagonal."""
    results = []
    for i in range(len(srcs)):
        for j in range(len(srcs)):
            if i != j:
                results.append(operator(srcs[i], srcs[j]))

    return results


def generate_results_empty(unused1, unused2):
    """Some tests don't need any explicit results stored in the shader."""
    return []


def abs_isub32(_a, _b):
    a = np.int32(np.uint32(_a))
    b = np.int32(np.uint32(_b))

    return np.uint32(a - b) if a > b else np.uint32(b - a)


def abs_isub64(_a, _b):
    a = np.int64(_a)
    b = _b.astype(np.int64)

    return np.uint64(a - b) if a > b else np.uint64(b - a)


def abs_usub32(_a, _b):
    a = np.uint32(_a)
    b = np.uint32(_b)

    return a - b if a > b else b - a


def abs_usub64(_a, _b):
    a = np.uint64(_a)
    b = np.uint64(_b)

    return a - b if a > b else b - a


def iadd_sat32(_a, _b):
    a = np.int32(np.uint32(_a))
    b = np.int32(np.uint32(_b))

    if a > 0:
        if b > (np.iinfo(np.int32).max - a):
            return np.iinfo(np.int32).max
    else:
        if b < (np.iinfo(np.int32).min - a):
            return np.iinfo(np.int32).min

    return a + b


def uadd_sat32(_a, _b):
    a = np.uint32(_a)
    b = np.uint32(_b)

    if b > (np.iinfo(np.uint32).max - a):
        return np.iinfo(np.uint32).max

    return a + b


def iadd_sat64(_a, _b):
    a = np.int64(_a)
    b = np.int64(_b)

    if a > 0:
        if b > (np.iinfo(np.int64).max - a):
            return np.iinfo(np.int64).max
    else:
        if b < (np.iinfo(np.int64).min - a):
            return np.iinfo(np.int64).min

    return a + b


def uadd_sat64(_a, _b):
    a = np.uint64(_a)
    b = np.uint64(_b)

    if b > (np.iinfo(np.uint64).max - a):
        return np.iinfo(np.uint64).max

    return a + b


def isub_sat32(a, b):
    r = np.int64(np.int32(a)) - np.int64(np.int32(b))

    if r > np.int64(0x07fffffff):
        return np.int32(0x7fffffff)

    if r < np.int64(-0x080000000):
        return np.int32(-0x80000000)

    return np.int32(r)


def usub_sat32(_a, _b):
    a = np.uint32(_a)
    b = np.uint32(_b)

    return a - b if a > b else np.uint32(0)


def isub_sat64(_a, _b):
    a = np.int64(_a)
    b = np.int64(_b)

    if a >= 0:
        if (a - np.iinfo(np.int64).max) > b:
            return np.iinfo(np.int64).max
    elif b >= 0:
        if a < (np.iinfo(np.int64).min + b):
            return np.iinfo(np.int64).min

    return a - b


def usub_sat64(_a, _b):
    a = np.uint64(_a)
    b = np.uint64(_b)

    return a - b if a > b else np.uint64(0)


def u_hadd32(_a, _b):
    a = np.uint32(_a)
    b = np.uint32(_b)

    return (a >> 1) + (b >> 1) + ((a & b) & 1)


def s_hadd32(_a, _b):
    a = np.int32(np.uint32(_a))
    b = np.int32(np.uint32(_b))

    return (a >> 1) + (b >> 1) + ((a & b) & 1)


def u_hadd64(_a, _b):
    a = np.uint64(_a)
    b = np.uint64(_b)

    return (a >> np.uint64(1)) + (b >> np.uint64(1)) + ((a & b) & np.uint64(1))


def s_hadd64(_a, _b):
    a = np.int64(_a)
    b = np.int64(_b)

    return (a >> np.int64(1)) + (b >> np.int64(1)) + ((a & b) & np.int64(1))


def u_rhadd32(_a, _b):
    a = np.uint32(_a)
    b = np.uint32(_b)

    return (a >> 1) + (b >> 1) + ((a | b) & 1)


def s_rhadd32(_a, _b):
    a = np.int32(np.uint32(_a))
    b = np.int32(np.uint32(_b))

    return (a >> 1) + (b >> 1) + ((a | b) & 1)


def u_rhadd64(_a, _b):
    a = np.uint64(_a)
    b = np.uint64(_b)

    return (a >> np.uint64(1)) + (b >> np.uint64(1)) + ((a | b) & np.uint64(1))


def s_rhadd64(_a, _b):
    a = np.int64(_a)
    b = np.int64(_b)

    return (a >> np.int64(1)) + (b >> np.int64(1)) + ((a | b) & np.int64(1))


def imul_32x16(a, b):
    return np.int32(a) * ((np.int32(b) << 16) >> 16)


def umul_32x16(a, b):
    return np.uint32(np.uint32(a) * (np.uint32(b) & 0x0000ffff))


def absoluteDifference32_sources():
    srcs = []
    for x in range(0, 32, 4):
        srcs += [ -(0x80000000 >> x), -(0x7fffffff >> x) ]

    srcs += [-5, -3, -1, 0, 1, 3, 5]

    for x in range(32 - 4, 0, -4):
        srcs += [ 0x7fffffff >> x, 0x80000000 >> x ]

    srcs.append(0x7fffffff)

    # Some prime numbers requiring from 14- to 32-bits to store.  The last is
    # actually negative.
    srcs += [ 0x00002ff9,
              0x0017fff5,
              0x017ffff5,
              0x05fffffb,
              0x2ffffff5,
              0xbffffff5
    ]

    return srcs


def absoluteDifference64_sources():
    srcs = []
    for x in range(0, 64, 6):
        srcs += [ -(0x8000000000000000 >> x), -(0x7fffffffffffffff >> x) ]

    srcs += [-5, -3, -2, -1, 0, 1, 2, 3, 5]

    for x in range(64 - 4, 0, -6):
        srcs += [ 0x7fffffffffffffff >> x, 0x8000000000000000 >> x ]

    srcs.append(0x7fffffffffffffff)

    # Some prime numbers requiring from 33- to 64-bits to store.  The last is
    # actually negative.
    srcs += [ 0x000000017ffffffb,    # 33 bits
              0x00000017ffffffef,    # 37 bits
              0x0000017ffffffff3,    # 41 bits
              0x000017ffffffffff,    # 45 bits
              0x00017fffffffffe1,    # 49 bits
              0x0005ffffffffffdd,    # 51 bits
              0x0017fffffffffff3,    # 53 bits
              0x017fffffffffffb5,    # 57 bits
              0x037fffffffffffe5,    # 58 bits
              0x17ffffffffffffe1,    # 61 bits
              0x5fffffffffffff89,    # 63 bits
              0xbfffffffffffffe1,    # 64 bits
    ]

    assert len(srcs) == 64
    return [np.uint64(x) for x in srcs]


def addSaturate_int32_sources():
    srcs = [0, 1, -1, 2, 3, 0x40000000, 0x7fffffff, -0x7fffffff, -0x80000000 ]

    random.seed(0)
    for i in range(2, 32):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    for i in range(4):
        srcs.append(random.randint(-0x7ffffffe, -2))

    assert len(srcs) == 43
    return srcs


def addSaturate_uint32_sources():
    srcs = [0, 1, 2, 3, 0x40000000, 0x7fffffff, 0x80000000, 0xf0f0f0f0, 0xff00ff00 ]

    random.seed(0)
    for i in range(2, 32):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    for i in range(43 - len(srcs)):
        srcs.append(random.randint(-0x7ffffffe, -2))

    assert len(srcs) == 43
    return srcs


def addSaturate_int64_sources():
    srcs = [0, 1, -1, 2, 3, 0x4000000000000000, 0x7fffffffffffffff, -0x7fffffffffffffff, -0x8000000000000000 ]

    random.seed(0)
    for i in range(16, 64):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    while len(srcs) < 62:
        srcs.append(random.randint(-0x7ffffffffffffffe, -2))

    assert len(srcs) == 62
    return [np.int64(np.uint64(x)) for x in srcs]


def addSaturate_uint64_sources():
    srcs = [0, 1, 2, 3, 0x4000000000000000, 0x7fffffffffffffff, 0x8000000000000000, 0xf0f0f0f0f0f0f0f0, 0xff00ff00ff00ff00 ]

    random.seed(0)
    for i in range(16, 64):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    while len(srcs) < 61:
        srcs.append(random.randint(0, 0xffffffffffffffff))

    srcs.append(np.uint64(0xdeadbeefdeadbeef))

    assert len(srcs) == 62
    return [np.uint64(x) for x in srcs]


def countLeadingZeros_sources():
    sources=[]
    random.seed(0)

    for i in range(1024):
        num_zeros = i % 33

        if i < 33:
            sources.append(0xffffffff >> num_zeros)
        else:
            sources.append((random.randint(0, 0xffffffff) | (1 << 31)) >> num_zeros)

    return sources


def countTrailingZeros_sources():
    sources=[]
    random.seed(0)

    for i in range(1024):
        num_zeros = i % 33

        if i < 33:
            sources.append(0xffffffff << num_zeros)
        else:
            sources.append((random.randint(0, 0xffffffff) | 1) << num_zeros)

    return sources


def multiply32x16_int32_sources():
    srcs = [0, 1, -1, np.int32(-0x80000000), -0x7fffffff, 0x7fffffff ]

    random.seed(0)
    for i in range(2, 32, 3):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    srcs.append(random.randint(0, 1 << 30) | (1 << 30))

    for i in range(2, 32, 3):
        srcs.append(-(random.randint(0, 1 << i) | (1 << i)))

    srcs.append(-(random.randint(0, 1 << 30) | (1 << 30)))

    while len(srcs) < 512:
        srcs.append(random.randint(-0x80000000, 0x7fffffff))

    return srcs


def subtractSaturate_int32_sources():
    srcs = [0, 1, -1, np.int32(-0x80000000), -0x7fffffff, 0x7fffffff ]

    random.seed(0)
    for i in range(2, 32, 3):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    srcs.append(random.randint(0, 1 << 30) | (1 << 30))

    for i in range(2, 32, 3):
        srcs.append(-(random.randint(0, 1 << i) | (1 << i)))

    srcs.append(-(random.randint(0, 1 << 30) | (1 << 30)))

    while len(srcs) < 32:
        srcs.append(random.randint(-0x80000000, 0x7fffffff))

    assert len(srcs) == 32
    return [np.int32(x) for x in srcs]


def subtractSaturate_uint32_sources():
    srcs = [0, 1, 0xf0f0f0f0 ]

    random.seed(0)
    for i in range(2, 31):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    assert len(srcs) == 32
    return srcs


def subtractSaturate_int64_sources():
    srcs = [0, 1, -1, -0x8000000000000000, -0x7fffffffffffffff, 0x7fffffffffffffff ]

    random.seed(0)
    for i in range(2, 32, 3):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    srcs.append(random.randint(0, 1 << 30) | (1 << 30))

    for i in range(16, 64, 3):
        srcs.append(-(random.randint(0, 1 << i) | (1 << i)))

    srcs.append(-(random.randint(0, 1 << 30) | (1 << 30)))

    while len(srcs) < 45:
        srcs.append(random.randint(-0x8000000000000000, 0x7fffffffffffffff))

    assert len(srcs) == 45
    return [np.int64(x) for x in srcs]


def subtractSaturate_uint64_sources():
    srcs = [0, 1, 0xf0f0f0f0f0f0f0f0 ]

    random.seed(0)
    for i in range(22, 64):
        srcs.append(random.randint(0, 1 << i) | (1 << i))

    assert len(srcs) == 45
    return srcs


FUNCS = {
    'absoluteDifference-int': {
        'input':      'int',
        'output':     'uint',
        'sources':    absoluteDifference32_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'absoluteDifference',
        'operator':   abs_isub32,
        'version':    '1.30',
        'extensions': None,
    },
    'absoluteDifference-uint': {
        'input':      'uint',
        'output':     'uint',
        'sources':    absoluteDifference32_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'absoluteDifference',
        'operator':   abs_usub32,
        'version':    '1.30',
        'extensions': None,
    },
    'absoluteDifference-int64': {
        'input':      'int64_t',
        'output':     'uint64_t',
        'sources':    absoluteDifference64_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'absoluteDifference',
        'operator':   abs_isub64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
    'absoluteDifference-uint64': {
        'input':      'uint64_t',
        'output':     'uint64_t',
        'sources':    absoluteDifference64_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'absoluteDifference',
        'operator':   abs_usub64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
    'countLeadingZeros-uint': {
        'input':      'uint',
        'output':     'uint',
        'sources':    countLeadingZeros_sources,
        'results':    generate_results_empty,
        'template':   'countLeadingZeros.shader_test.mako',
        'func':       'countLeadingZeros',
        'operator':   None,
        'version':    '1.30',
        'extensions': None,
    },
    'countTrailingZeros-uint': {
        'input':      'uint',
        'output':     'uint',
        'sources':    countTrailingZeros_sources,
        'results':    generate_results_empty,
        'template':   'countLeadingZeros.shader_test.mako',
        'func':       'countTrailingZeros',
        'operator':   None,
        'version':    '1.30',
        'extensions': None,
    },
    'addSaturate-int': {
        'input':      'int',
        'output':     'int',
        'sources':    addSaturate_int32_sources,
        'results':    generate_results_commutative_with_diagonal,
        'template':   'addSaturate.shader_test.mako',
        'func':       'addSaturate',
        'operator':   iadd_sat32,
        'version':    '1.30',
        'extensions': None,
    },
    'addSaturate-uint': {
        'input':      'uint',
        'output':     'uint',
        'sources':    addSaturate_uint32_sources,
        'results':    generate_results_commutative_with_diagonal,
        'template':   'addSaturate.shader_test.mako',
        'func':       'addSaturate',
        'operator':   uadd_sat32,
        'version':    '1.30',
        'extensions': None,
    },
    'addSaturate-int64': {
        'input':      'int64_t',
        'output':     'int64_t',
        'sources':    addSaturate_int64_sources,
        'results':    generate_results_commutative_with_diagonal,
        'template':   'addSaturate.shader_test.mako',
        'func':       'addSaturate',
        'operator':   iadd_sat64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
    'addSaturate-uint64': {
        'input':      'uint64_t',
        'output':     'uint64_t',
        'sources':    addSaturate_uint64_sources,
        'results':    generate_results_commutative_with_diagonal,
        'template':   'addSaturate.shader_test.mako',
        'func':       'addSaturate',
        'operator':   uadd_sat64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
    'average-int': {
        'input':      'int',
        'output':     'int',
        'sources':    absoluteDifference32_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'average',
        'operator':   s_hadd32,
        'version':    '1.30',
        'extensions': None,
    },
    'average-uint': {
        'input':      'uint',
        'output':     'uint',
        'sources':    absoluteDifference32_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'average',
        'operator':   u_hadd32,
        'version':    '1.30',
        'extensions': None,
    },
    'average-int64': {
        'input':      'int64_t',
        'output':     'int64_t',
        'sources':    absoluteDifference64_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'average',
        'operator':   s_hadd64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
    'average-uint64': {
        'input':      'uint64_t',
        'output':     'uint64_t',
        'sources':    absoluteDifference64_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'average',
        'operator':   u_hadd64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
        },
    'averageRounded-int': {
        'input':      'int',
        'output':     'int',
        'sources':    absoluteDifference32_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'averageRounded',
        'operator':   s_rhadd32,
        'version':    '1.30',
        'extensions': None,
    },
    'averageRounded-uint': {
        'input':      'uint',
        'output':     'uint',
        'sources':    absoluteDifference32_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'averageRounded',
        'operator':   u_rhadd32,
        'version':    '1.30',
        'extensions': None,
    },
    'averageRounded-int64': {
        'input':      'int64_t',
        'output':     'int64_t',
        'sources':    absoluteDifference64_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'averageRounded',
        'operator':   s_rhadd64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
    'averageRounded-uint64': {
        'input':      'uint64_t',
        'output':     'uint64_t',
        'sources':    absoluteDifference64_sources,
        'results':    generate_results_commutative,
        'template':   'absoluteDifference.shader_test.mako',
        'func':       'averageRounded',
        'operator':   u_rhadd64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
    'multiply32x16-int': {
        'input':      'int',
        'output':     'int',
        'sources':    multiply32x16_int32_sources,
        'results':    generate_results_empty,
        'template':   'multiply32x16.shader_test.mako',
        'func':       'multiply32x16',
        'operator':   imul_32x16,
        'version':    '1.30',
        'extensions': None,
    },
    'multiply32x16-uint': {
        'input':      'uint',
        'output':     'uint',
        'sources':    multiply32x16_int32_sources,
        'results':    generate_results_empty,
        'template':   'multiply32x16.shader_test.mako',
        'func':       'multiply32x16',
        'operator':   umul_32x16,
        'version':    '1.30',
        'extensions': None,
    },
    'subtractSaturate-int': {
        'input':      'int',
        'output':     'int',
        'sources':    subtractSaturate_int32_sources,
        'results':    generate_results_without_diagonal,
        'template':   'subtractSaturate.shader_test.mako',
        'func':       'subtractSaturate',
        'operator':   isub_sat32,
        'version':    '1.30',
        'extensions': None,
    },
    'subtractSaturate-uint': {
        'input':      'uint',
        'output':     'uint',
        'sources':    subtractSaturate_uint32_sources,
        'results':    generate_results_without_diagonal,
        'template':   'subtractSaturate.shader_test.mako',
        'func':       'subtractSaturate',
        'operator':   usub_sat32,
        'version':    '1.30',
        'extensions': None,
    },
    'subtractSaturate-int64': {
        'input':      'int64_t',
        'output':     'int64_t',
        'sources':    subtractSaturate_int64_sources,
        'results':    generate_results_without_diagonal,
        'template':   'subtractSaturate.shader_test.mako',
        'func':       'subtractSaturate',
        'operator':   isub_sat64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
    'subtractSaturate-uint64': {
        'input':      'uint64_t',
        'output':     'uint64_t',
        'sources':    subtractSaturate_uint64_sources,
        'results':    generate_results_without_diagonal,
        'template':   'subtractSaturate.shader_test.mako',
        'func':       'subtractSaturate',
        'operator':   usub_sat64,
        'version':    '4.00',  # GL_ARB_gpu_shader_int64 requires 4.0.
        'extensions': 'GL_ARB_gpu_shader_int64',
    },
}


def main():
    """main function."""
    dirname = os.path.join('spec', 'intel_shader_integer_functions2',
                           'execution', 'built-in-functions')
    utils.safe_makedirs(dirname)

    for func, attrib in six.iteritems(FUNCS):

        TEMPLATE = template_file(os.path.basename(os.path.splitext(__file__)[0]),
                                 attrib['template'])

        for execution_stage in ('vs', 'fs'):
            filename = os.path.join(
                dirname, "{0}-{1}.shader_test".format(execution_stage, func))
            print(filename)

            extension_list = ["GL_INTEL_shader_integer_functions2"]
            if isinstance(attrib['extensions'], str):
                extension_list += [attrib['extensions']]
            elif attrib['extensions'] is not None:
                extension_list += attrib['extensions']

            with open(filename, 'w') as f:
                f.write(TEMPLATE.render_unicode(
                    execution_stage=execution_stage,
                    version=attrib['version'],
                    extensions=sorted(extension_list),
                    input_type=attrib['input'],
                    output_type=attrib['output'],
                    sources=attrib['sources'](),
                    results=attrib['results'](attrib['sources'](), attrib['operator']),
                    func=attrib['func']
                ))
    return

if __name__ == '__main__':
    main()



#!/usr/bin/env python2
# coding=utf-8

# Copyright (c) 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import errno
import random_ubo

def do_test(requirements, packing):
    path = os.path.join("spec", "arb_uniform_buffer_object", "execution")

    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise
        pass

    basename = random_ubo.generate_file_name(requirements, packing)
    fullname = os.path.join(path, basename)

    file = open(fullname, "w", 0)

    fields, required_layouts = random_ubo.generate_ubo(
        requirements,
        random_ubo.ALL130_TYPES)

    layouts = random_ubo.generate_layouts(
        fields,
        required_layouts,
        # Due to bugs in the NVIDIA closed-source driver, do not randomly
        # generate layout(row_major) on structures.  Several tests will,
        # however, do this explicitly.
        False)

    blocks = random_ubo.generate_block_list(
        130,
        packing,
        fields,
        layouts)

    print basename
    file.write(random_ubo.emit_shader_test(
        blocks,
        packing,
        130,
        ["GL_ARB_uniform_buffer_object","GL_ARB_arrays_of_arrays"]))

    file.close()

all_packing = [random_ubo.std140_packing_rules(),
               random_ubo.shared_packing_rules()]

all_requirements = []

# Generate a test for each matrix type that:
#
# - Explicitly declares one as row-major and another as column-major.
#
# - Embeds the matrix in a structure without a layout.
#
# - Embeds the matrix in a structure with a row-major layout.
#
# - Embeds the matrix in a structure with a column-major layout.
#
# - Each of the above in an array

for m in ["mat2x2", "mat2x3", "mat2x4",
          "mat3x2", "mat3x3", "mat3x4",
          "mat4x2", "mat4x3", "mat4x4"]:
    all_requirements.append([["row_major", m], ["column_major", m]])
    all_requirements.append([["#column_major", "struct", m]])
    all_requirements.append([["row_major",     "struct", m]])
    all_requirements.append([["column_major",  "struct", m]])

    all_requirements.append([["row_major", "array", m],
                             ["column_major", "array", m]])
    all_requirements.append([["#column_major", "struct", "array", m]])
    all_requirements.append([["row_major",     "struct", "array", m]])
    all_requirements.append([["column_major",  "struct", "array", m]])

    all_requirements.append([["#column_major", "array", "struct", m]])
    all_requirements.append([["row_major",     "array", "struct", m]])
    all_requirements.append([["column_major",  "array", "struct", m]])

    all_requirements.append([["#column_major", "array", "struct", "array", m]])
    all_requirements.append([["row_major",     "array", "struct", "array", m]])
    all_requirements.append([["column_major",  "array", "struct", "array", m]])

# Also add some struct-nesting tests.

all_requirements.append([["struct", "struct"]])
all_requirements.append([["struct", "struct", "struct"]])
all_requirements.append([["struct", "array", "struct"]])
all_requirements.append([["array", "struct", "struct"]])
all_requirements.append([["array", "struct", "array", "struct"]])

all_requirements.append([["struct", "array", "array", "struct"]])
all_requirements.append([["struct", "array", "array", "array", "struct"]])
all_requirements.append([["array", "array", "struct", "array"]])
all_requirements.append([["struct", "array", "array", "array"]])

for p in all_packing:
    for r in all_requirements:
        do_test(r, p)



# coding=utf-8
# Copyright 2014-2016,2018 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import abc
import os
import subprocess

import six
from six.moves import range

from framework import core, grouptools, exceptions
from framework import options
from framework.profile import TestProfile
from framework.test.base import Test, is_crash_returncode, TestRunError

__all__ = [
    'DEQPBaseTest',
    'gen_caselist_txt',
    'get_option',
    'iter_deqp_test_cases',
    'make_profile',
]


def get_option(env_varname, config_option, default=None, required=False):
    """Query the given environment variable and then piglit.conf for the option.

    Return the value of the default argument if opt is None.

    """
    opt = os.environ.get(env_varname, None)
    if opt is not None:
        return opt

    opt = core.PIGLIT_CONFIG.safe_get(config_option[0], config_option[1])

    if required and not opt:
        raise exceptions.PiglitFatalError(
            'Cannot get env "{}" or conf value "{}:{}"'.format(
                env_varname, config_option[0], config_option[1]))
    return opt or default


_EXTRA_ARGS = get_option('PIGLIT_DEQP_EXTRA_ARGS',
                         ('deqp', 'extra_args'),
                         default='').split()


def select_source(bin_, filename, mustpass, extra_args):
    """Return either the mustpass list or the generated list."""
    if options.OPTIONS.deqp_mustpass:
        return gen_mustpass_tests(mustpass)
    else:
        return iter_deqp_test_cases(
            gen_caselist_txt(bin_, filename, extra_args))


def make_profile(test_list, test_class):
    """Create a TestProfile instance."""
    profile = TestProfile()
    for testname in test_list:
        # deqp uses '.' as the testgroup separator.
        piglit_name = testname.replace('.', grouptools.SEPARATOR)
        profile.test_list[piglit_name] = test_class(testname)

    return profile


def gen_mustpass_tests(mustpass):
    """Return a testlist from the mustpass list."""
    with open(mustpass, 'r') as f:
        for l in f:
            l = l.strip()
            if l:
                yield l


def gen_caselist_txt(bin_, caselist, extra_args):
    """Generate a caselist.txt and return its path.

    Extra args should be a list of extra arguments to pass to deqp.

    """
    # dEQP is stupid (2014-12-07):
    #   1. To generate the caselist file, dEQP requires that the process's
    #      current directory must be that same as that of the executable.
    #      Otherwise, it fails to find its data files.
    #   2. dEQP creates the caselist file in the process's current directory
    #      and provides no option to change its location.
    #   3. dEQP creates a GL context when generating the caselist. Therefore,
    #      the caselist must be generated on the test target rather than the
    #      build host. In other words, when the build host and test target
    #      differ then we cannot pre-generate the caselist on the build host:
    #      we must *dynamically* generate it during the testrun.
    basedir = os.path.dirname(bin_)
    caselist_path = os.path.join(basedir, caselist)

    # TODO: need to catch some exceptions here...
    with open(os.devnull, 'w') as d:
        env = os.environ.copy()
        env['MESA_GL_VERSION_OVERRIDE'] = '4.6'
        env['MESA_GLES_VERSION_OVERRIDE'] = '3.2'

        subprocess.check_call(
            [bin_, '--deqp-runmode=txt-caselist'] + extra_args, cwd=basedir,
            stdout=d, stderr=d, env=env)
    assert os.path.exists(caselist_path)
    return caselist_path


def iter_deqp_test_cases(case_file):
    """Iterate over original dEQP testcase names."""
    with open(case_file, 'r') as caselist_file:
        for i, line in enumerate(caselist_file):
            if line.startswith('GROUP:'):
                continue
            elif line.startswith('TEST:'):
                yield line[len('TEST:'):].strip()
            else:
                raise exceptions.PiglitFatalError(
                    'deqp: {}:{}: ill-formed line'.format(case_file, i))


@six.add_metaclass(abc.ABCMeta)
class DEQPBaseTest(Test):
    __RESULT_MAP = {
        "Pass": "pass",
        "Fail": "fail",
        "QualityWarning": "warn",
        "InternalError": "fail",
        "Crash": "crash",
        "NotSupported": "skip",
        "ResourceError": "crash",
    }

    @abc.abstractproperty
    def deqp_bin(self):
        """The path to the exectuable."""

    @abc.abstractproperty
    def extra_args(self):
        """Extra arguments to be passed to the each test instance.

        Needs to return a list, since self.command uses the '+' operator, which
        only works to join two lists together.

        """
        return _EXTRA_ARGS

    def __init__(self, case_name):
        command = [self.deqp_bin, '--deqp-case=' + case_name]

        super(DEQPBaseTest, self).__init__(command)

        # dEQP's working directory must be the same as that of the executable,
        # otherwise it cannot find its data files (2014-12-07).
        # This must be called after super or super will overwrite it
        self.cwd = os.path.dirname(self.deqp_bin)

    @Test.command.getter
    def command(self):
        """Return the command plus any extra arguments."""
        command = super(DEQPBaseTest, self).command
        return command + self.extra_args

    def __find_map(self):
        """Run over the lines and set the result."""
        # splitting this into a separate function allows us to return cleanly,
        # otherwise this requires some break/else/continue madness
        for line in self.result.out.split('\n'):
            line = line.lstrip()
            for k, v in six.iteritems(self.__RESULT_MAP):
                if line.startswith(k):
                    self.result.result = v
                    return

    def interpret_result(self):
        if is_crash_returncode(self.result.returncode):
            self.result.result = 'crash'
        elif self.result.returncode != 0:
            self.result.result = 'fail'
        else:
            self.__find_map()

        # We failed to parse the test output. Fallback to 'fail'.
        if self.result.result == 'notrun':
            self.result.result = 'fail'

    def _run_command(self, *args, **kwargs):
        """Rerun the command if X11 connection failure happens."""
        for _ in range(5):
            super(DEQPBaseTest, self)._run_command(*args, **kwargs)
            x_err_msg = "FATAL ERROR: Failed to open display"
            if x_err_msg in self.result.err or x_err_msg in self.result.out:
                continue
            return

        raise TestRunError('Failed to connect to X server 5 times', 'fail')



# coding=utf-8
#
# Copyright © 2012, 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import print_function, division, absolute_import
import os

from six.moves import range

from templates import template_dir
from modules import utils

TEMPLATES = template_dir(os.path.splitext(os.path.basename(__file__))[0])

# These are a set of pseudo random values used by the number sequence
# generator.  See get_value above.
RANDOM_NUMBERS = (0.78685, 0.89828, 0.36590, 0.92504, 0.48998, 0.27989,
                  0.08693, 0.48144, 0.87644, 0.18080, 0.95147, 0.18892,
                  0.45851, 0.76423, 0.78659, 0.97998, 0.24352, 0.60922,
                  0.45241, 0.33045, 0.27233, 0.92331, 0.63593, 0.67826,
                  0.12195, 0.24853, 0.35977, 0.41759, 0.79119, 0.54281,
                  0.04089, 0.03877, 0.58445, 0.43017, 0.58635, 0.48151,
                  0.58778, 0.37033, 0.47464, 0.17470, 0.18308, 0.49466,
                  0.45838, 0.30337, 0.71273, 0.45083, 0.88339, 0.47350,
                  0.86539, 0.48355, 0.92923, 0.79107, 0.77266, 0.71677,
                  0.79860, 0.95149, 0.05604, 0.16863, 0.14072, 0.29028,
                  0.57637, 0.13572, 0.36011, 0.65431, 0.38951, 0.73245,
                  0.69497, 0.76041, 0.31016, 0.48708, 0.96677, 0.58732,
                  0.33741, 0.73691, 0.24445, 0.35686, 0.72645, 0.65438,
                  0.00824, 0.00923, 0.87650, 0.43315, 0.67256, 0.66939,
                  0.87706, 0.73880, 0.96248, 0.24148, 0.24126, 0.24673,
                  0.18999, 0.10330, 0.78826, 0.23209, 0.59548, 0.23134,
                  0.72414, 0.88036, 0.54498, 0.32668, 0.02967, 0.12643)

ALL_TEMPLATES = ("",
                 "-from-const",
                 "-set-by-API",
                 "-set-by-other-stage")


def get_value(type_, idx):
    """Get a string representing a number in the specified GLSL type"""

    value = RANDOM_NUMBERS[idx % len(RANDOM_NUMBERS)]

    if type_[0] == 'b':
        if (value * 10) > 5:
            return "1"
        else:
            return "0"
    elif type_[0] == 'i':
        return str(int(value * 100) - 50)
    elif type_[0] == 'u':
        return str(int(value * 50))
    else:
        return str((value * 20.0) - 10.0)


def generate_tests(type_list, base_name, major, minor):
    dirname = os.path.join('spec',
                           'glsl-{0}.{1}'.format(major, minor),
                           'execution',
                           'uniform-initializer')
    utils.safe_makedirs(dirname)

    for target in ("vs", "fs"):
        for t in ALL_TEMPLATES:
            template = TEMPLATES.get_template(
                "{0}-initializer{1}.shader_test.mako".format(target, t))

            test_file_name = os.path.join(
                dirname,
                '{0}-{1}{2}.shader_test'.format(target, base_name, t))
            print(test_file_name)

            # Generate the test vectors.  This is a list of tuples.  Each
            # tuple is a type name paired with a value.  The value is
            # formatted as a GLSL constructor.
            #
            # A set of types and values is also generated that can be set via
            # the OpenGL API.  Some of the tests use this information.
            test_vectors = []
            api_vectors = []
            for i, (type_, num_values) in enumerate(type_list):
                numbers = []
                alt_numbers = []
                for j in range(num_values):
                    numbers.append(get_value(type_, i + j))
                    alt_numbers.append(get_value(type_, i + j + 7))

                value = "{0}({1})".format(type_, ", ".join(numbers))

                api_type = type_
                if type_ == "bool":
                    api_type = "int"
                elif type_[0] == 'b':
                    api_type = "ivec{0}".format(type_[-1])

                if type_[-1] in ["2", "3", "4"]:
                    name = 'u{0}{1}'.format(type_[0], type_[-1])
                else:
                    name = 'u{0}'.format(type_[0])

                test_vectors.append((type_, name, value))
                api_vectors.append((api_type, name, alt_numbers))

            with open(test_file_name, "w") as f:
                f.write(template.render_unicode(type_list=test_vectors,
                                                api_types=api_vectors,
                                                major=major,
                                                minor=minor))


def generate_array_tests(type_list, base_name, major, minor):
    dirname = os.path.join('spec',
                           'glsl-{0}.{1}'.format(major, minor),
                           'execution',
                           'uniform-initializer')
    utils.safe_makedirs(dirname)

    def parts():
        """Generate parts."""
        # pylint: disable=undefined-loop-variable
        for j in range(2):
            numbers = []
            for k in range(num_values):
                numbers.append(get_value(type_, i + j + k))

            yield '{0}({1})'.format(type_, ', '.join(numbers))
        # pylint: enable=undefined-loop-variable

    vecs = []
    for i, (type_, num_values) in enumerate(type_list):
        if type_[-1] in ["2", "3", "4"]:
            name = 'u{0}{1}'.format(type_[0], type_[-1])
        else:
            name = 'u{0}'.format(type_[0])

        array_type = '{0}[2]'.format(type_)
        value = "{0}({1})".format(array_type, ", ".join(parts()))

        vecs.append((array_type, name, value))

    for target in ("vs", "fs"):
        template = TEMPLATES.get_template(
            '{0}-initializer.shader_test.mako'.format(target))

        test_file_name = os.path.join(
            dirname,
            '{0}-{1}-array.shader_test'.format(target, base_name))
        print(test_file_name)

        with open(test_file_name, "w") as f:
            f.write(template.render_unicode(type_list=vecs,
                                            major=major,
                                            minor=minor))


def main():
    """Main function."""
    bool_types = [("bool", 1), ("bvec2", 2), ("bvec3", 3), ("bvec4", 4)]
    int_types = [("int", 1), ("ivec2", 2), ("ivec3", 3), ("ivec4", 4)]
    float_types = [("float", 1), ("vec2", 2), ("vec3", 3), ("vec4", 4)]
    uint_types = [("uint", 1), ("uvec2", 2), ("uvec3", 3), ("uvec4", 4)]
    mat2_types = [("mat2x2", 4), ("mat2x3", 6), ("mat2x4", 8)]
    mat3_types = [("mat3x2", 6), ("mat3x3", 9), ("mat3x4", 12)]
    mat4_types = [("mat4x2", 8), ("mat4x3", 12), ("mat4x4", 16)]

    for types, base_name, major, minor in [(bool_types, "bool", 1, 20),
                                           (int_types, "int", 1, 20),
                                           (float_types, "float", 1, 20),
                                           (mat2_types, "mat2", 1, 20),
                                           (mat3_types, "mat3", 1, 20),
                                           (mat4_types, "mat4", 1, 20),
                                           (uint_types, "uint", 1, 30)]:
        generate_tests(types, base_name, major, minor)
        generate_array_tests(types, base_name, major, minor)


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright (c) 2015-2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the oglconform integration."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
try:
    import mock
except ImportError:
    from unittest import mock

import pytest
from six.moves import cStringIO as StringIO

from framework import grouptools
from framework import status

with mock.patch('framework.core.PIGLIT_CONFIG.required_get',
                mock.Mock(return_value='piglit.conf.example')):
    with mock.patch('subprocess.call', mock.Mock()):
        from tests import oglconform

# pylint: disable=protected-access,no-self-use


class TestMakeProfile(object):
    """Tests for the _make_profile function."""

    def test_basic(self, mocker):
        """tests.oglconform._make_profile: Adds test names"""
        io_ = mocker.Mock(wraps=StringIO(u'group test.name\n'))
        io_.name = mocker.Mock()

        mock_file = mocker.MagicMock()
        mock_file.__enter__.return_value = io_

        mock_temp = mocker.patch.object(oglconform.tempfile,
                                        'NamedTemporaryFile')
        mock_temp.return_value = mock_file

        mocker.patch('subprocess.call', mocker.Mock())
        profile = oglconform._make_profile()

        name = grouptools.join('oglconform', 'group', 'test.name')
        assert name in profile.test_list

    def test_missing(self, mocker):
        """tests.oglconform._make_profile: handles missing groups"""
        io_ = mocker.Mock(wraps=StringIO(u'test.name\n'))
        io_.name = mocker.Mock()

        mock_file = mocker.MagicMock()
        mock_file.__enter__.return_value = io_

        mock_temp = mocker.patch.object(oglconform.tempfile,
                                        'NamedTemporaryFile')
        mock_temp.return_value = mock_file

        mocker.patch('subprocess.call', mocker.Mock())
        oglconform._make_profile()


class TestOGLCTest(object):
    """Tests for the OGLCTest class."""

    def test_command(self):
        """tests.oglconform.OGLCtest.command: value is as expected"""
        expected = ['piglit.conf.example', '-minFmt', '-v', '4', '-test',
                    'group', 'test']

        test = oglconform.OGLCTest('group', 'test')
        assert expected == test.command

    class TestInterpretResult(object):
        """Tests for the interpret_result method."""

        def test_pass(self):
            """tests.oglconform.OGLCtest.interpret_result: status is pass when
            expected.
            """
            test = oglconform.OGLCTest('group', 'test')
            test.result.returncode = 0
            test.result.out = (
                'Another line\n'
                'Total Passed : 1\n'
                'Total Failed : 0\n'
                'Total Not run: 0\n'
            )
            test.interpret_result()

            assert test.result.result is status.PASS

        def test_skip_from_not_run(self):
            """tests.oglconform.OGLCtest.interpret_result: status is skip when
            tests not run.
            """
            test = oglconform.OGLCTest('group', 'test')
            test.result.returncode = 0
            test.result.out = (
                'Another line\n'
                'Total Failed : 0\n'
                'Total Passed : 0\n'
                'Total Not run: 1\n'
            )
            test.interpret_result()

            assert test.result.result is status.SKIP

        def test_fail(self):
            """tests.oglconform.OGLCtest.interpret_result: status is fail
            otherwise.
            """
            test = oglconform.OGLCTest('group', 'test')
            test.result.returncode = 0
            test.interpret_result()

            assert test.result.result is status.FAIL

        def test_crash(self):
            """tests.oglconform.OGLCtest.interpret_result: status is crash if
            returncode is not 0.
            """
            test = oglconform.OGLCTest('group', 'test')
            test.result.returncode = -1
            test.interpret_result()

            assert test.result.result is status.CRASH

    @pytest.mark.parametrize("out", [
        'no test in schedule is compat',
        'GLSL 1.30 is not supported',
        'GLSL 1.40 is not supported',
        'GLSL 1.50 is not supported',
        'GLSL 3.30 is not supported',
        'GLSL 3.40 is not supported',
        'GLSL 3.50 is not supported',
        'wont be scheduled due to lack of compatible fbconfig',
    ])
    def test_skip_from_re(self, out):
        test = oglconform.OGLCTest('group', 'value')
        test.result.out = out
        test.result.returncode = 0
        test.interpret_result()
        assert test.result.result == 'skip'



# encoding=utf-8
# Copyright © 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Generate tests for extensions that affect the shading language.

This generates three tests per extensions.
- The first test asserts that if the extension is required it's defined
- The second test asserts that if the extension isn't available it's not
  defined
- The third test asserts that if the extension isn't available trying to
  require it is an error.

These are glslparsertests.

If the requirement for the test is less than OpenGL Shader Language 140, then
two tests will be made, a core and a compat. If it's greater then only a core
will be made, these will have compat and core added to the name, respectively.

For OpenGL ES only one test will be generated, it will have es added to the
name of the test.

"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import os

from templates import template_dir
from modules import utils, glsl

_TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))
ENABLED_TEMPLATE = _TEMPLATES.get_template('enabled.glsl.mako')
DISABLED_TEMPLATE = _TEMPLATES.get_template('disabled.glsl.mako')
UNDEFINED_TEMPLATE = _TEMPLATES.get_template('undefined-require.glsl.mako')

# A list of tuples with the full preprocess defined name, and the minimum
# supported version of that extension.
EXTENSIONS = [
    ("GL_ARB_draw_instanced", "110"),
    ("GL_ARB_draw_buffers", "110"),
    ("GL_ARB_enhanced_layouts", "140"),
    ("GL_ARB_separate_shader_objects", "110"),
    ("GL_ARB_texture_rectangle", "110"),
    ("GL_AMD_shader_trinary_minmax", "110"),
    ("GL_EXT_texture_array", "110"),
    ("GL_ARB_ES3_1_compatibility", "430"),
    ("GL_ARB_arrays_of_arrays", "110"),
    ("GL_ARB_fragment_coord_conventions", "110"),
    ("GL_ARB_fragment_layer_viewport", "150"),
    ("GL_ARB_explicit_attrib_location", "110"),
    ("GL_ARB_explicit_uniform_location", "110"),
    ("GL_ARB_shader_texture_lod", "110"),
    ("GL_AMD_conservative_depth", "110"),
    ("GL_ARB_conservative_depth", "110"),
    ("GL_ARB_shader_bit_encoding", "110"),
    ("GL_ARB_shader_clock", "110"),
    ("GL_ARB_uniform_buffer_object", "110"),
    ("GL_ARB_texture_cube_map_array", "110"),
    ("GL_ARB_shading_language_packing", "110"),
    ("GL_ARB_texture_multisample", "110"),
    ("GL_ARB_texture_query_levels", "110"),
    ("GL_ARB_texture_query_lod", "110"),
    ("GL_ARB_gpu_shader5", "150"),
    ("GL_ARB_gpu_shader_fp64", "150"),
    ("GL_ARB_vertex_attrib_64bit", "150"),
    ("GL_AMD_vertex_shader_layer", "130"),
    ("GL_AMD_vertex_shader_viewport_index", "110"),
    ("GL_ARB_shading_language_420pack", "110"),
    ("GL_ARB_sample_shading", "110"),
    ("GL_ARB_texture_gather", "110"),
    ("GL_ARB_shader_atomic_counters", "110"),
    ("GL_ARB_shader_atomic_counter_ops", "140"),
    ("GL_ARB_viewport_array", "110"),
    ("GL_ARB_compute_shader", "110"),
    ("GL_ARB_shader_image_load_store", "130"),
    ("GL_ARB_shader_image_size", "110"),
    ("GL_ARB_shader_texture_image_samples", "110"),
    # That is what the original hand written test required.
    ("GL_ARB_derivative_control", "150"),
    ("GL_ARB_shader_precision", "110"),
    ("GL_ARB_shader_storage_buffer_object", "110"),
    ("GL_ARB_tessellation_shader", "150"),
    ("GL_ARB_shader_subroutine", "150"),
    ("GL_ARB_shader_draw_parameters", "140"),
    ("GL_EXT_separate_shader_objects", "100"),
    ("GL_EXT_draw_buffers", "100"),
    ("GL_AMD_shader_stencil_export", "120"),
    ("GL_ARB_shader_stencil_export", "120"),
    ("GL_ARB_geometry_shader4", "110"),
    ("GL_OES_EGL_image_external", "100"),
    ("GL_EXT_shader_samples_identical", "110"),
    ("GL_ARB_shader_group_vote", "110"),
    ("GL_EXT_shader_samples_identical", "310 es"),
    ("GL_OES_sample_variables", "300 es"),
    ("GL_OES_multisample_interpolation", "300 es"),
    ("GL_OES_standard_derivatives", "100"),
    ("GL_OES_texture_storage_multisample_2d_array", "310 es"),
    ("GL_OES_blend_func_extended", "100"),
    ("GL_OES_shader_image_atomic", "310 es"),
    ("GL_OES_shader_io_blocks", "310 es"),
    ("GL_EXT_shader_io_blocks", "310 es"),
    ("GL_OES_geometry_shader", "310 es"),
    ("GL_EXT_geometry_shader", "310 es"),
    ("GL_OES_geometry_point_size", "310 es"),
    ("GL_EXT_geometry_point_size", "310 es"),
    ("GL_EXT_gpu_shader5", "310 es"),
    ("GL_OES_gpu_shader5", "310 es"),
    ("GL_EXT_texture_buffer", "310 es"),
    ("GL_OES_texture_buffer", "310 es"),
    ("GL_EXT_clip_cull_distance", "300 es"),
    ("GL_NV_image_formats", "310 es"),
]
EXTENSIONS = [(n, glsl.Version(v)) for n, v in EXTENSIONS]


def _gen_tests(ext, version, stage, path, extra_name, extra_extensions):
    """Generate the actual test files.

    This generates both a disabled test for both disabled and enabled
    configurations.

    Arguments:
    ext -- the extension to be tested
    version -- the minimum required version
    stage -- the shader stage to test
    path -- the path to put the test in, without the test name
    extra_name -- An extra string to put in the test name.
    extra_extensions -- Any extra extensions requires.

    """
    for test, template in [('enabled', ENABLED_TEMPLATE),
                           ('disabled-defined', DISABLED_TEMPLATE),
                           ('disabled-undefined', UNDEFINED_TEMPLATE)]:
        name = os.path.join(path, '{}-{}.{}'.format(test, extra_name, stage))

        # Open in bytes mode to avoid weirdness in python 2/3 compatibility
        with open(name, 'wb') as f:
            f.write(template.render(
                version=version,
                extension=ext,
                extra_extensions=extra_extensions))
        print(name)


def main():
    """Main function."""
    for ext, ver in EXTENSIONS:
        # Lower ext in the path name, but keep it capitalized for the generated
        # tests
        path = os.path.join('spec', ext[3:].lower(), 'preprocessor')
        utils.safe_makedirs(path)

        for stage in ['vert', 'frag', 'geom', 'tesc', 'tese', 'comp']:
            # Calculate the minimum version for the stage, with extensions.
            # This makes the maximum number of tests run on the widest swath of
            # hardware.
            version, extra_ext = glsl.MinVersion.for_stage_with_ext(stage, ver)
            if extra_ext is not None:
                extra_extensions = [extra_ext]
            else:
                extra_extensions = []

            if not version.is_es:
                # if the actual version is less than 140 make a compat test and
                # a core test, otherwise just make a core test
                if version < 140:
                    _gen_tests(ext, version, stage, path, 'compat',
                               extra_extensions)
                    _gen_tests(ext, glsl.Version('140'), stage, path, 'core',
                               extra_extensions)
                else:
                    _gen_tests(ext, version, stage, path, 'core',
                               extra_extensions)
            else:
                _gen_tests(ext, version, stage, path, 'es', extra_extensions)


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Copyright © 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


"""Generate a set of shader_runner tests for overloaded versions of every
 built-in function specified in arb_shader_precision, based on the test
 vectors computed by builtin_function.py, and structured according to the mako
 templates in templates/gen_shader_precision_tests.

 The vertex, geometry, and fragment shader types are exercised by each test.
 In all cases, the inputs to the built-in functions come from uniforms, so
 that the effectiveness of the test won't be circumvented by constant folding
 in the GLSL compiler.

 The tests operate by invoking the built-in function in the appropriate
 shader, calculating any deviance from the expected value (in ulps), comparing
 the deviance to a supplied tolerance (according to those specified in
 arb_shader_precision), and then outputting the pass/fail result as a solid
 rgba color which is then checked using shader_runner's "probe rgba" command.

 For built-in functions whose result type is multi-valued (vec or mat), the
 tests calculate the error in ulps for each element separately, but compare
 only the largest error value to the tolerance.  This accounts for cases where
 error varies among the elements of a vec or mat result.

 This program updates the tolerance values in the test vectors with tolerances
 sourced from the GLSL spec (v4.5) and adds a few new tests as well.

 This program outputs, to stdout, the name of each file it generates.
"""

from __future__ import print_function, division, absolute_import
from builtin_function import *
import os
import numpy
import struct

import six
from six.moves import range

from templates import template_file



tolerances = {'pow': 16.0,
              'exp': 3.0,
              'exp2': 3.0,
              'log': 3.0,
              'log2': 3.0,
              'sqrt': 3.0,
              'inversesqrt': 2.0,
              'op-div': 2.5,
              'op-assign-div': 2.5,
              }

shader_precision_spec_fns = ('op-add', 'op-assign-add',
                             'op-sub', 'op-assign-sub',
                             'op-mult', 'op-assign-mult',
                             'op-div', 'op-assign-div',
                             'degrees', 'radians',
                             'pow',
                             'exp',
                             'exp2',
                             'log', 'log2',
                             'sqrt', 'inversesqrt')

def _is_sequence(arg):
    try:
        # python 3.3+
        Sequence = collections.abc.Sequence
    except AttributeError:
        # python 2.7+, removed in 3.9
        Sequence = collections.Sequence
    return isinstance(arg, (Sequence, numpy.ndarray))

def _floatToBits(f):
    s = struct.pack('>f', f)
    return struct.unpack('>L', s)[0]

# The precision for log and log2 from the GLSL spec:
# 3 ULP outside the range [0.5, 2.0].
# Absolute error < 2^-21 inside the range [0.5, 2.0].
def _log_tol(compcnt, args, i, j):
    arg = args[j][i] if compcnt > 1 else args[0]
    if arg < 0.5 or arg > 2.0:
        return 3.0
    elif arg >= 0.5 and arg < 1.0:
        # in the range [0.5, 1.0), one ulp is 5.960464478e-08,
        # so 2^-21 is 8 ulps
        return 8.0
    else:
        # in the range [1.0, 2.0), one ulp is 1.192092896e-07,
        # so 2^-21 is 4 ulps
        return 4.0

# The precision for exp and exp2 from the GLSL spec:
# (3 + 2 * |x|) ULP.
def _exp_tol(compcnt, args, i, j):
    arg = args[j][i] if compcnt > 1 else args[0]
    return 3.0 + 2.0 * abs(arg)

# The precision for division from the GLSL spec:
# 2.5 ULP for b in the range [2^-126, 2^126].
def _div_tol(compcnt, args, i, j):
    divisor = args[1][i] if _is_sequence(args[1]) else args[1]
    assert (divisor >= pow(2,-126) and divisor <= pow(2,126)) or \
           (divisor <= -pow(2,-126) and divisor >= -pow(2,126))
    return 2.5

# The precision for pow from the GLSL spec:
# Inherited from exp2 (x * log2 (y)).
def _pow_tol(compcnt, args, i, j):
    x = args[0][i] if _is_sequence(args[0]) else args[0]
    y = args[1][i] if _is_sequence(args[1]) else args[1]
    return(_exp_tol(0, [x * _log_tol(0, [y], 0, 0)], 0, 0))


_tol_fns = { 'op-add': (lambda *args: 0.0),
             'op-assign-add': (lambda *args: 0.0),
             'op-sub': (lambda *args: 0.0),
             'op-assign-sub': (lambda *args: 0.0),
             'op-mult': (lambda *args: 0.0),
             'op-assign-mult': (lambda *args: 0.0),
             'op-div': (lambda compcnt, args, i, j: _div_tol(compcnt, args, i, j)),
             'op-assign-div': (lambda compcnt, args, i, j: _div_tol(compcnt, args, i, j)),
             'degrees': (lambda *args: 2.5),
             'radians': (lambda *args: 2.5),
             'pow': (lambda compcnt, args, i, j: _pow_tol(compcnt, args, i, j)),
             'exp': _exp_tol,
             'exp2': _exp_tol,
             'log': _log_tol,
             'log2': _log_tol,
             'sqrt': (lambda *args: 2.5),
             'inversesqrt': (lambda *args: 2.0) }


def _gen_tolerance(name, rettype, args):
    ret = []
    compcnt = rettype.num_cols * rettype.num_rows

    for j in range(rettype.num_cols):
        for i in range(rettype.num_rows):
            assert _tol_fns[name]
            ret.append(_tol_fns[name](compcnt, args, i, j))
    return ret

def make_indexers(signature):
   """Build a list of strings which index into every possible
   value of the result.  For example, if the result is a vec2,
   then build the indexers ['[0]', '[1]'].
   """
   if signature.rettype.num_cols == 1:
      col_indexers = ['']
   else:
      col_indexers = ['[{0}]'.format(i) for i in range(signature.rettype.num_cols)]
   if signature.rettype.num_rows == 1:
      row_indexers = ['']
   else:
      row_indexers = ['[{0}]'.format(i) for i in range(signature.rettype.num_rows)]
   return [col_indexer + row_indexer
           for col_indexer in col_indexers for row_indexer in row_indexers]

def shader_runner_type(glsl_type):
    """Return the appropriate type name necessary for binding a
    uniform of the given type using shader_runner's "uniform" command.
    Boolean values and vectors are converted to ints, and square
    matrices are written in "matNxN" form.
    """
    if glsl_type.base_type == glsl_bool:
        if glsl_type.is_scalar:
            return 'int'
        else:
            return 'ivec{0}'.format(glsl_type.num_rows)
    elif glsl_type.is_matrix:
        return 'mat{0}x{1}'.format(glsl_type.num_cols, glsl_type.num_rows)
    else:
        return str(glsl_type)

def shader_runner_format(values):
    """Format the given values for use in a shader_runner "uniform" or
    "probe rgba" command.  Sequences of values are separated by spaces.
    """

    if _is_sequence(values):
        retval = ''
        for x in values:
            assert isinstance(x, (float, np.float32))
            retval += ' {0:#08x}'.format(_floatToBits(x))
    else:
        assert isinstance(values, (float, np.float32))
        retval = '{0:#08x}'.format(_floatToBits(values))

    return retval

def drop_signbit(xs):
    if (np.isscalar(xs)):
        return xs.dtype.type(0.0) if xs == 0.0 else xs
    return np.array([xs.dtype.type(0.0) if x == 0.0 else x for x in xs],
                    xs.dtype)

def main():
    """ Main function """

    for signature, test_vectors in six.iteritems(test_suite):
        arg_float_check = all(arg.base_type == glsl_float for arg in signature.argtypes)
        arg_mat_check = any(arg.is_matrix for arg in signature.argtypes)
        # Filter the test vectors down to only those which deal exclusively in
        # non-matrix float types and are specified in the spec
        if (signature.rettype.base_type == glsl_float and
            arg_float_check and
            signature.name in shader_precision_spec_fns and
            not arg_mat_check):
            # replace the tolerances in each test_vector with
            # our own tolerances specified in ulps
            refined_test_vectors = []
            complex_tol_type = signature.rettype
            for test_vector in test_vectors:
                tolerance = _gen_tolerance(signature.name, signature.rettype, test_vector.arguments)
                result = drop_signbit(test_vector.result)
                refined_test_vectors.append(TestVector(test_vector.arguments, result, tolerance))
            # Then generate the shader_test scripts
            for shader_stage in ('vs', 'fs', 'gs'):
                template = template_file('gen_shader_precision_tests', '{0}.mako'.format(shader_stage))
                output_filename = os.path.join( 'spec', 'arb_shader_precision',
                                                '{0}-{1}-{2}.shader_test'.format(
                                                shader_stage, signature.name,
                                                '-'.join(str(argtype)
                                                for argtype in signature.argtypes)))
                print(output_filename)
                dirname = os.path.dirname(output_filename)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                indexers = make_indexers(signature)
                num_elements = signature.rettype.num_cols * signature.rettype.num_rows
                invocation = signature.template.format( *['arg{0}'.format(i)
                                                        for i in range(len(signature.argtypes))])
                with open(output_filename, 'w') as f:
                    f.write(template.render_unicode( signature=signature,
                                                     is_complex_tolerance=_is_sequence(tolerance),
                                                     complex_tol_type=signature.rettype,
                                                     test_vectors=refined_test_vectors,
                                                     invocation=invocation,
                                                     num_elements=num_elements,
                                                     indexers=indexers,
                                                     shader_runner_type=shader_runner_type,
                                                     shader_runner_format=shader_runner_format,
                                                     column_major_values=column_major_values ))

if __name__ == "__main__":
    main()
